// =======================================================================
// CBP + FBO = MAS22 Simulation Framework
// MAS22_AirTrack_MassSpring.cpp
// Template for Matt Pethybridge
// 01-07-22
// 07-07-22
// =======================================================================

#include "MAS22_AirTrack_MassSpring.h"
#include "MAS22_Actor.h"
#include "MAS22_Structs.h"
#include "MAS22_Solver_RK.h"
#include "MAS22_Solver_RKF.h"
#include "Components/CapsuleComponent.h"
#include "Components/StaticMeshComponent.h"

// 24(C) Define indices into the parameter array. The order must agree with
// order of adding into the parameter array
#define INDEX_MASS 0
#define INDEX_SPRINGCONSTANT 1
#define INDEX_DAMPING 2
// -----------------------------------------------------------------------

// 27(C) Define indices into the initial conditions array. The order must
//  agree with order of adding into the array
#define INDEX_INITDISPX 0
#define INDEX_INITVELYX 1
// -----------------------------------------------------------------------

// 28(C) Number of equations in rhs --------------------------------------
#define NREQUNS 2
// -----------------------------------------------------------------------

AMAS22_AirTrack_MassSpring::AMAS22_AirTrack_MassSpring() {

	PrimaryActorTick.bCanEverTick = false;

	CapsuleCollider = CreateDefaultSubobject<UCapsuleComponent>("CapsuleCollider");
	CapsuleCollider->SetupAttachment(GetRootComponent());
	CapsuleCollider->SetCapsuleHalfHeight(200);
	CapsuleCollider->SetCapsuleRadius(100);
	CapsuleCollider->SetRelativeLocation(FVector(0, 0, 0));

	// 1(C) Put in code to generate air track ----------------------------

	static ConstructorHelpers::FObjectFinder<UStaticMesh>MeshAsset1(TEXT("StaticMesh'/Game/Air_Track/AirTrackShort.AirTrackShort'"));
	UStaticMesh* Asset1 = MeshAsset1.Object;

	Track = CreateDefaultSubobject<UStaticMeshComponent>("Track");
	Track->SetStaticMesh(Asset1);
	Track->SetRelativeLocation(FVector(0, 0, 0));
	Track->SetupAttachment(CapsuleCollider);

	// 3(C) Put in code to generate the glider ---------------------------
	static ConstructorHelpers::FObjectFinder<UStaticMesh>MeshAsset2(TEXT("StaticMesh'/Game/Air_Track/AirTrackGlider.AirTrackGlider'"));
	UStaticMesh* Asset2 = MeshAsset2.Object;

	Glider = CreateDefaultSubobject<UStaticMeshComponent>("Glider1");
	Glider->SetStaticMesh(Asset2);
	Glider->SetRelativeLocation(FVector(0, 0, 0));
	Glider->SetupAttachment(CapsuleCollider);

	// 5(C) Put in code to generate the Left spring ----------------------
	static ConstructorHelpers::FObjectFinder<UStaticMesh>MeshAsset3(TEXT("StaticMesh'/Game/Air_Track/Spring_UVs_LOD_0.Spring_UVs_LOD_0'"));
	UStaticMesh* Asset3 = MeshAsset3.Object;

	LeftSpring = CreateDefaultSubobject<UStaticMeshComponent>("LeftSpring");
	LeftSpring->SetStaticMesh(Asset3);
	LeftSpring->SetRelativeLocation(FVector(-140, 0, 55));  // 7(C) Tweak location of left spring
	LeftSpring->SetupAttachment(CapsuleCollider);

	// 8(C) Put in code to generate the Right spring ----------------------
	RightSpring = CreateDefaultSubobject<UStaticMeshComponent>("RightSpring");
	RightSpring->SetStaticMesh(Asset3);
	RightSpring->SetRelativeLocation(FVector(140, 0, 55));  // 7(C) Tweak location of left spring
	RightSpring->SetupAttachment(CapsuleCollider);

	// 10(C) Put in code to generate the left stop -----------------------
	static ConstructorHelpers::FObjectFinder<UStaticMesh>MeshAsset4(TEXT("StaticMesh'/Game/Air_Track/AirTrackStop.AirTrackStop'"));
	UStaticMesh* Asset4 = MeshAsset4.Object;

	LeftStop = CreateDefaultSubobject<UStaticMeshComponent>("LeftStop");
	LeftStop->SetStaticMesh(Asset4);
	LeftStop->SetRelativeLocation(FVector(-279, 0, 0));  // 12(C) Tweak location of left stop
	LeftStop->SetupAttachment(CapsuleCollider);

	// 13(C) Put in code to generate the right stop -----------------------
	RightStop = CreateDefaultSubobject<UStaticMeshComponent>("RightStop");
	RightStop->SetStaticMesh(Asset4);
	RightStop->SetRelativeLocation(FVector(279, 0, 0));
	RightStop->SetRelativeRotation(FRotator(0, 180, 0).Quaternion());  // 15(C) Need to tweak the relative rotation
	RightStop->SetupAttachment(CapsuleCollider);

	// 32(C) get the glider start location -------------------------------
	gliderStartLocation = Glider->GetRelativeLocation();
	// -------------------------------------------------------------------

	// 35(C) get the springs start location ------------------------------
	leftSpringStartLocation  = LeftSpring->GetRelativeLocation();
	rightSpringStartLocation = RightSpring->GetRelativeLocation();
	// -------------------------------------------------------------------

}

void AMAS22_AirTrack_MassSpring::ActorInitialize() { //-------------------
	Super::ActorInitialize();

	// config file stuff moved to function -------------------------------
	ApplyConfig();

	// 09-07-22 Establish global logging
	InitGlobalLog();
	flog("Hi from AMAS22_AirTrack_MassSpring\n");

	u = new double[NREQUNS];
	dudt = new double[NREQUNS];
	yout = new double[NREQUNS]; // Needed for RK

	SetModelICs();
	UpdateModelICs();

	GetDefaultParameters();
	SetModelParameters();
	UpdateModelParameters();

	InitializeSolver();

	// Create the numeric solvers ---------------------------------------
	solverRK = new MAS22_Solver_RK();
	solverRK->init(NREQUNS);

	solverRKF = new MAS22_Solver_RKF();

	// 29(C) Create the RHS for this model -------------------------------
	rightHandSide = new rhs_AirTrack_MK(massP,springConstantP,dampingP);
	// -------------------------------------------------------------------

}

// Deal with parameters --------------------------------------------------

void AMAS22_AirTrack_MassSpring::SetModelParameters() {
	modelParametersStruct line;
	modelParameters.Empty();

	// 22(C) Build up parameter array ------------------------------------
	line.val = massP;
	line.name = "mass";
	modelParameters.Add(line);

	line.val = springConstantP;
	line.name = "springConstant";
	modelParameters.Add(line);

	line.val = dampingP;
	line.name = "damping";
	modelParameters.Add(line);

}

void AMAS22_AirTrack_MassSpring::UpdateModelParameters() {
	
	//23(C) get updated values of parameters if menu changes them --------
	massP = modelParameters[INDEX_MASS].val;
	springConstantP = modelParameters[INDEX_SPRINGCONSTANT].val;
	dampingP = modelParameters[INDEX_DAMPING].val;
	// -------------------------------------------------------------------

	// Below added 06-07-22
		// - Change the params so we need a new rhs
	delete rightHandSide;
	//printf("UpdateModelParameters() Before create rightHandSide = %p", rightHandSide);
	rightHandSide = new rhs_AirTrack_MK(massP, springConstantP, dampingP);
	//printf("UpdateModelParameters() After create rightHandSide = %p", rightHandSide);
}

// added 07-07-22
void AMAS22_AirTrack_MassSpring::GetDefaultParameters() {
	dampingP = damping;
	massP = mass;
	springConstantP = springConstant;
}

// Model initial conditions ----------------------------------------------

void AMAS22_AirTrack_MassSpring::SetModelICs() {
	Super::SetModelICs();

	modelParametersStruct icLine;
	modelInitialConditions.Empty();			// allows for hard reset

	// 25(C) Build up IC array -------------------------------------------
	icLine.val = initDispX;
	icLine.name = "init DispX";
	modelInitialConditions.Add(icLine);

	icLine.val = initVelyX;
	icLine.name = "init VelyX";
	modelInitialConditions.Add(icLine);
	// -------------------------------------------------------------------
}

void AMAS22_AirTrack_MassSpring::UpdateModelICs() {

	// 26(C) get updated values of ICs if menu changes them --------------
	initDispX = modelInitialConditions[INDEX_INITDISPX].val;
	initVelyX = modelInitialConditions[INDEX_INITVELYX].val;
	// -------------------------------------------------------------------
	 
	// last line must be
	InitializeSolver(); // to set the ICs
}

void AMAS22_AirTrack_MassSpring::InitializeSolver() {
	flag = 2;
	time = 0.0;
	relerr = 10 * sqrt(solverRKF->r8_epsilon());
	abserr = 10 * sqrt(solverRKF->r8_epsilon());

	printf("AMAS22_AirTrack_MassSpring::InitializeSolver() relerr = %e", relerr);

	// 29(C) Assign variables to initial conditions ----------------------
	u[0] = initDispX;
	u[1] = initVelyX;

	// set variable names from u[] array
	// 06-07-22
	dispX = u[0];
	velyX = u[1];

}

void AMAS22_AirTrack_MassSpring::BuildHUDLines() {
	HUDLine line;

	HUDLines.Empty();
	bHUD_exists = true;
	if (bHUD_exists) {
		Super::BuildHUDLines();
	}

	// 39(C) Add additional HUD lines for this model ---------------------
	line.message = FString::Printf(TEXT("dispX = %f"), dispX);
	line.colour = VARIBLE_COLOR;
	line.value = 0;
	HUDLines.Add(line);

	line.message = FString::Printf(TEXT("velyX = %f"), velyX);
	line.colour = VARIBLE_COLOR;
	line.value = 0;
	HUDLines.Add(line);
	// -------------------------------------------------------------------

}

void AMAS22_AirTrack_MassSpring::LogData() {
	int m;
	m = (int(logInterval / dT));
	if (itn % (m+1) == 0) {		// m changed to m+1
		dataRecords.Add(FDataRecord_ATMS{ time, dispX, velyX });
	}
}


// This is the main 'loop' function called on an Unreal Timer ------------
// Don't add any model-specific code in here
void AMAS22_AirTrack_MassSpring::ComputeDT() {
	if (bRunning) {
		LogData();					// 06-07-22 Moved to capture initial conds
		Computation();
		Visualization(dT);
		BuildHUDLines();
		itn++;
	}
}

void AMAS22_AirTrack_MassSpring::Computation() {

	switch (solver) {
	case SolverEnum::EULER:
		Euler(dT);
		break;
	case SolverEnum::RUNGE_KUTTA:
		RK(dT);
	case SolverEnum::RUNGE_KUTTA_FEHLBERG:
		RKF(dT);
		break;
	default:
		break;
	}

	// Addit2 (C) set model variables from u[] array ---------------------
	dispX = u[0];
	velyX = u[1];

	// -------------------------------------------------------------------
	// 30(C) Output computations for debugging ---------------------------
	//printf("time = %f, dispX = %f, velyX = %f\n", time, dispX, velyX);
	// -------------------------------------------------------------------

}

// Numerical solver code -------------------------------------------------

void AMAS22_AirTrack_MassSpring::Euler(double deltaT) {

}

void AMAS22_AirTrack_MassSpring::RK(double deltaT) {
	// Following for the Runge-Kutta experimental ------------------------
	double h = deltaT;
    flag = solverRK->step(u, dudt, time, h, yout, *rightHandSide);
    for (int i = 0; i < NREQUNS; i++) u[i] = yout[i];
	time += deltaT;
}

void AMAS22_AirTrack_MassSpring::RKF(double deltaT) {

	tout = time + deltaT;

	//printf(">>>> PRE  time = %f, tout = %f", time, tout);

	flag = solverRKF->r8_rkf45(NREQUNS, u, dudt, &time, tout, &relerr, abserr, flag, *rightHandSide);

	//printf(">>>> POST time = %f, tout = %f", time, tout);

	if (flag == 3) {
		printErr("Flag = 3 reset");
		flag = 2;
	}
	if (flag == 4) {
		//	printErr("Flag = 4 reset");
		flag = 2;
	}
	if (flag == 5) {
		printErr("Flag = 5 exiting");
		exit(0);
	}
	if (flag == 6) {
		printErr("Flag = 6 reset");
		flag = 2;
	}
	if (flag == 7) {
		//	printErr("Flag = 7 reset");
		flag = 2;
	}

}


void AMAS22_AirTrack_MassSpring::Visualization(float deltaTime) { // -------------------
	
	// 33(C) Move the glider ---------------------------------------------
	FVector newLocation = gliderStartLocation;
	newLocation.X += scaling * dispX;
	Glider->SetRelativeLocation(newLocation);
	// -------------------------------------------------------------------

	// 36(C) Move the springs --------------------------------------------
	newLocation = leftSpringStartLocation;
	newLocation.X += (scaling * dispX)/2.0;
	LeftSpring->SetRelativeLocation(newLocation);
	newLocation = rightSpringStartLocation;
	newLocation.X += (scaling * dispX)/2.0;
	RightSpring->SetRelativeLocation(newLocation);
	// -------------------------------------------------------------------

	// 38(C) Scale the springs -------------------------------------------
	FVector scaleVec;
	scaleVec = LeftSpring->GetRelativeScale3D();
	scaleVec.X = (springLength + (dispX * scaling)) / springLength;
	LeftSpring->SetRelativeScale3D(scaleVec);

	scaleVec = LeftSpring->GetRelativeScale3D();
	scaleVec.X = (springLength - (dispX * scaling)) / springLength;
	RightSpring->SetRelativeScale3D(scaleVec);
	// -------------------------------------------------------------------
}

// Console commands ------------------------------------------------------

void AMAS22_AirTrack_MassSpring::SetConsoleAddit(FString name, FString strVal) {

}

void AMAS22_AirTrack_MassSpring::DumpConsoleAll() {
	Super::DumpConsoleAll();

}

// Create Octave file here -----------------------------------------------

void AMAS22_AirTrack_MassSpring::EndPlay(const EEndPlayReason::Type EndPlayReason) {
	Super::EndPlay(EndPlayReason);

	if(solver == SolverEnum::RUNGE_KUTTA) solverRK->cleanup();

	// Here comes the Octave code -------------

	FString logFile = FPaths::ProjectDir();
	FString fileName = "OctaveFiles/";
	fileName.Append(actorName);
	fileName.Append(".m");
	logFile.Append(fileName);

	int nrRecords = dataRecords.Num();

	printf("EndPlay() nrRecords = %d\n", nrRecords);

	FString stringToSave = "%col1 time, col2 dispX, col3 velyX\n\n";

	// Declare the Octave matrix "A" -------------------------------------
	stringToSave.Append("A=[\n");

	// Add the data from dataRecords[...] into array "A" -----------------
	for (int i = 0; i < nrRecords; i++) {
		stringToSave.Append(FString::Printf(TEXT("%e,  "), dataRecords[i].time));
		stringToSave.Append(FString::Printf(TEXT("%e,  "), dataRecords[i].dispX));
		stringToSave.Append(FString::Printf(TEXT("%e,\n"), dataRecords[i].velyX));
	}

	// Close the Octave matrix "A" ---------------------------------------
	stringToSave.Append("];\n\n");

	// Create some Octave variable names, from matrix "A" ----------------
	stringToSave.Append("time   = A(:,1);\n");
	stringToSave.Append("dispX  = A(:,2);\n");
	stringToSave.Append("velyX  = A(:,3);\n");

	// Get the parameters into Octave ------------------------------------
	stringToSave.Append(logf("m = %f, k = %f, b = %f\n", massP, springConstantP,dampingP));
	stringToSave.Append(logf("initDispX = %f, initVelyX = %f\n", initDispX, initVelyX));
	stringToSave.Append("if(initDispX == 0.0) initDispX = 1e-10; endif;");

	// calc the theoretical solution --------------------------------------
	stringToSave.Append("tau = 2*m/(b + 1e-7);\n");
	stringToSave.Append("omega = sqrt(k/m - (b/(2*m))^2);\n");
	stringToSave.Append("t=initDispX^2 + (1/omega^2)*(initVelyX + initDispX/tau)^2;\n");
	stringToSave.Append(" A = sqrt(t);\n");
	//stringToSave.Append("phi = -asin( (initVelyX + initDispX/tau)/(omega*A));;\n");
	stringToSave.Append("phi = -atan((1/tau + initVelyX/initDispX)/omega);\n");

	stringToSave.Append("xTh = A*exp(-time/tau).*cos(omega*time + phi);\n");
	stringToSave.Append("vTh = -(1/tau)*A*exp(-time/tau).*cos(omega*time+phi) - ...;\n");
	stringToSave.Append("A*exp(-time/tau)*omega.*sin(omega*time + phi);\n");

	// Plot experimental and theoretical solution ------------------------
	// First displacement ------------------------------------------------

	stringToSave.Append("figure(1)\n");
	stringToSave.Append("plot(time,dispX,'o',time,xTh,'linewidth',3);\n");
	stringToSave.Append(logf("title('m = %f, k = %f, b = %f');\n", massP, springConstantP, dampingP));
	stringToSave.Append("xlabel('time')\n");
	stringToSave.Append("ylabel('dispX')\n");
	stringToSave.Append(logf("grid;\n"));

	// Then velocity -----------------------------------------------------

	stringToSave.Append("figure(2)\n");
	stringToSave.Append("plot(time,velyX,'o',time,vTh,'linewidth',3);\n");
	stringToSave.Append(logf("title('m = %f, k = %f, b = %f');\n", massP, springConstantP, dampingP));
	stringToSave.Append("xlabel('time')\n");
	stringToSave.Append("ylabel('velyX')\n");
	stringToSave.Append(logf("grid;\n"));

	// All done. Write file to disk -------------------------------------

	FFileHelper::SaveStringToFile(stringToSave, *logFile);

}

void AMAS22_AirTrack_MassSpring::ApplyConfig() {
	// config file stuff --------------------------
	if (editorOverride == true) {
		FString configFName = FPaths::GeneratedConfigDir().Append(TEXT("MAS22_System.ini"));
		//printf("AMAS22_AirTrack_MassSpring::ActorInitialize() path to config file = %s", *configFName);
		GConfig->LoadFile(configFName);

		// System defaults -------------------------------------------------------------------------

		GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_AirTrack_MassSpring"), TEXT("scaling"), scaling, configFName);
		printf("AMAS22_AirTrack_MassSpring::ActorInitialize() >>>>>Config File read: scaling set to %f", scaling);

		GConfig->GetString(TEXT("/Script/MAS22_SciencePark.MAS22_AirTrack_MassSpring"), TEXT("defaultSolver"), defaultSolver, configFName);
		printf("AMAS22_AirTrack_MassSpring::ActorInitialize() >>>>>Config File read: solver set to %s", *defaultSolver);
		if (defaultSolver == "RUNGE_KUTTA_FEHLBERG") solver = SolverEnum::RUNGE_KUTTA_FEHLBERG;

		GConfig->GetString(TEXT("/Script/MAS22_SciencePark.MAS22_AirTrack_MassSpring"), TEXT("actorName"), actorName, configFName);
		printf("AMAS22_AirTrack_MassSpring::ActorInitialize() >>>>>Config File read: actor Name set to %s", *actorName);

		GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_AirTrack_MassSpring"), TEXT("dT"), dT, configFName);
		printf("AMAS22_AirTrack_MassSpring::ActorInitialize() >>>>>Config File read: dT set to %f", dT);

		// Default parameters ----------------------------------------------------------------------

		GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_AirTrack_MassSpring"), TEXT("mass"), mass, configFName);
		printf("AMAS22_AirTrack_MassSpring::ActorInitialize() >>>>>Config File read: default_mass set to %f", mass);

		GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_AirTrack_MassSpring"), TEXT("springConstant"), springConstant, configFName);
		printf("AMAS22_AirTrack_MassSpring::ActorInitialize() >>>>>Config File read: springConstant set to %f", springConstant);

		GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_AirTrack_MassSpring"), TEXT("damping"), damping, configFName);
		printf("AMAS22_AirTrack_MassSpring::ActorInitialize() >>>>>Config File read: damping set to %f", damping);
	}
}