// =======================================================================
// CBP + FBO = MAS22 Simulation Framework
// MAS22_UTube.cpp
// Template for Matt Pethybridge
// 01-07-22
// 07-07-22
// =======================================================================

#include "MAS22_UTube.h"
#include "MAS22_Actor.h"
#include "MAS22_Structs.h"
#include "MAS22_Solver_RK.h"
#include "MAS22_Solver_RKF.h"
#include "Components/CapsuleComponent.h"
#include "Components/StaticMeshComponent.h"

// 24(C) Define indices into the parameter array. The order must agree with
// order of adding into the parameter array
#define INDEX_GRAVY 0
#define INDEX_TOTALLENGTH 1
#define INDEX_FRICTION 2
// -----------------------------------------------------------------------

// 27(C) Define indices into the initial conditions array. The order must
//  agree with order of adding into the array
#define INDEX_INITDISPZ 0
#define INDEX_INITVELYZ 1
// -----------------------------------------------------------------------

// 28(C) Number of equations in rhs --------------------------------------
#define NREQUNS 2
// -----------------------------------------------------------------------

AMAS22_UTube::AMAS22_UTube() {

	PrimaryActorTick.bCanEverTick = false;

	CapsuleCollider = CreateDefaultSubobject<UCapsuleComponent>("CapsuleCollider");
	CapsuleCollider->SetupAttachment(GetRootComponent());
	CapsuleCollider->SetCapsuleHalfHeight(200);
	CapsuleCollider->SetCapsuleRadius(100);
	CapsuleCollider->SetRelativeLocation(FVector(0, 0, 0));

	// 1(C) Put in code to generate Utube --------------------------------

	static ConstructorHelpers::FObjectFinder<UStaticMesh>MeshAsset1(TEXT("StaticMeshStaticMesh'/Game/UTube/UTube.UTube'"));
	UStaticMesh* Asset1 = MeshAsset1.Object;

	UTube = CreateDefaultSubobject<UStaticMeshComponent>("UTube");
	UTube->SetStaticMesh(Asset1);
	UTube->SetRelativeLocation(FVector(0, 0, 0));
	UTube->SetupAttachment(CapsuleCollider);

	// 3(C) Put in code to generate the glider ---------------------------
	static ConstructorHelpers::FObjectFinder<UStaticMesh>MeshAsset2(TEXT("StaticMeshStaticMesh'/Game/UTube/ULiquid.ULiquid'"));
	UStaticMesh* Asset2 = MeshAsset2.Object;

	ULiquid = CreateDefaultSubobject<UStaticMeshComponent>("ULiquid1");
	ULiquid->SetStaticMesh(Asset2);
	ULiquid->SetRelativeLocation(FVector(0, 0, 0));
	ULiquid->SetupAttachment(CapsuleCollider);

	// 5(C) Put in code to generate the Left spring ----------------------
	static ConstructorHelpers::FObjectFinder<UStaticMesh>MeshAsset3(TEXT("StaticMeshStaticMesh'/Game/UTube/ULiquidColumn.ULiquidColumn'"));
	UStaticMesh* Asset3 = MeshAsset3.Object;

	LeftColumn = CreateDefaultSubobject<UStaticMeshComponent>("LeftColumn");
	LeftColumn->SetStaticMesh(Asset3);
	LeftColumn->SetRelativeLocation(FVector(-100, 0, 100));  // 7(C) Tweak location of left Column
	LeftColumn->SetupAttachment(CapsuleCollider);

	// 8(C) Put in code to generate the Right Column ----------------------
	RightColumn = CreateDefaultSubobject<UStaticMeshComponent>("RightColumn");
	RightColumn->SetStaticMesh(Asset3);
	RightColumn->SetRelativeLocation(FVector(100, 0, 100));  // 7(C) Tweak location of left Column
	RightColumn->SetupAttachment(CapsuleCollider);

	// 10(C) Put in code to generate the left stop -----------------------
	static ConstructorHelpers::FObjectFinder<UStaticMesh>MeshAsset4(TEXT("StaticMeshStaticMesh'/Game/UTube/Stand.Stand'"));
	UStaticMesh* Asset4 = MeshAsset4.Object;

	Stand = CreateDefaultSubobject<UStaticMeshComponent>("Stand");
	Stand->SetStaticMesh(Asset4);
	Stand->SetRelativeLocation(FVector(-279, 0, 0));  // 12(C) Tweak location of left stop
	Stand->SetupAttachment(CapsuleCollider);

	static ConstructorHelpers::FObjectFinder<UStaticMesh>MeshAsset5(TEXT("StaticMeshStaticMeshStaticMesh'/Game/UTube/Clamp.Clamp'"));
	UStaticMesh* Asset5 = MeshAsset5.Object;

	// 13(C) Put in code to generate the right stop -----------------------
	Clamp = CreateDefaultSubobject<UStaticMeshComponent>("Clamp");
	Clamp->SetStaticMesh(Asset5);
	Clamp->SetRelativeLocation(FVector(279, 0, 0));
	Clamp->SetRelativeRotation(FRotator(0, 180, 0).Quaternion());  // 15(C) Need to tweak the relative rotation
	Clamp->SetupAttachment(CapsuleCollider);

	// 32(C) get the glider start location -------------------------------
	//gliderStartLocation = Glider->GetRelativeLocation();
	// -------------------------------------------------------------------

	// 35(C) get the springs start location ------------------------------
	//leftSpringStartLocation = LeftSpring->GetRelativeLocation();
	//rightSpringStartLocation = RightSpring->GetRelativeLocation();
	// -------------------------------------------------------------------

}

void AMAS22_UTube::ActorInitialize() { //-------------------
	Super::ActorInitialize();

	// config file stuff moved to function -------------------------------
	ApplyConfig();

	// 09-07-22 Establish global logging
	InitGlobalLog();
	flog("Hi from AMAS22_UTube\n");

	u = new double[NREQUNS];
	dudt = new double[NREQUNS];
	yout = new double[NREQUNS]; // Needed for RK

	SetModelICs();
	UpdateModelICs();

	GetDefaultParameters();
	SetModelParameters();
	UpdateModelParameters();

	InitializeSolver();

	// Create the numeric solvers ---------------------------------------
	solverRK = new MAS22_Solver_RK();
	solverRK->init(NREQUNS);

	solverRKF = new MAS22_Solver_RKF();

	// 29(C) Create the RHS for this model -------------------------------
	rightHandSide = new rhs_UTube(gravityP, totalLengthP, frictionCoeffP);
	// -------------------------------------------------------------------

}

// Deal with parameters --------------------------------------------------

void AMAS22_UTube::SetModelParameters() {
	modelParametersStruct line;
	modelParameters.Empty();

	// 22(C) Build up parameter array ------------------------------------
	line.val = gravityP;
	line.name = "gravity";
	modelParameters.Add(line);

	line.val = totalLengthP;
	line.name = "totalLength";
	modelParameters.Add(line);

	line.val = frictionCoeffP;
	line.name = "frictionCoeff";
	modelParameters.Add(line);

}

void AMAS22_UTube::UpdateModelParameters() {

	//23(C) get updated values of parameters if menu changes them --------
	gravityP = modelParameters[INDEX_GRAVY].val;
	totalLengthP = modelParameters[INDEX_TOTALLENGTH].val;
	frictionCoeffP = modelParameters[INDEX_FRICTION].val;
	// -------------------------------------------------------------------

	// Below added 06-07-22
		// - Change the params so we need a new rhs
	delete rightHandSide;
	//printf("UpdateModelParameters() Before create rightHandSide = %p", rightHandSide);
	rightHandSide = new rhs_UTube(gravityP, totalLengthP, frictionCoeffP);
	//printf("UpdateModelParameters() After create rightHandSide = %p", rightHandSide);
}

// added 07-07-22
void AMAS22_UTube::GetDefaultParameters() {
	gravityP = gravity;
	totalLengthP = totalLength;
	frictionCoeffP = frictionCoeff;
}

// Model initial conditions ----------------------------------------------

void AMAS22_UTube::SetModelICs() {
	Super::SetModelICs();

	modelParametersStruct icLine;
	modelInitialConditions.Empty();			// allows for hard reset

	// 25(C) Build up IC array -------------------------------------------
	icLine.val = initDispZ;
	icLine.name = "init DispZ";
	modelInitialConditions.Add(icLine);

	icLine.val = initVelyZ;
	icLine.name = "init VelyZ";
	modelInitialConditions.Add(icLine);
	// -------------------------------------------------------------------
}

void AMAS22_UTube::UpdateModelICs() {

	// 26(C) get updated values of ICs if menu changes them --------------
	initDispZ = modelInitialConditions[INDEX_INITDISPZ].val;
	initVelyZ = modelInitialConditions[INDEX_INITVELYZ].val;
	// -------------------------------------------------------------------

	// last line must be
	InitializeSolver(); // to set the ICs
}

void AMAS22_UTube::InitializeSolver() {
	flag = 2;
	time = 0.0;
	relerr = 10 * sqrt(solverRKF->r8_epsilon());
	abserr = 10 * sqrt(solverRKF->r8_epsilon());

	printf("AMAS22_UTube::InitializeSolver() relerr = %e", relerr);

	// 29(C) Assign variables to initial conditions ----------------------
	u[0] = initDispZ;
	u[1] = initVelyZ;

	// set variable names from u[] array
	// 06-07-22
	dispZ = u[0];
	velyZ = u[1];

}

void AMAS22_UTube::BuildHUDLines() {
	HUDLine line;

	HUDLines.Empty();
	bHUD_exists = true;
	if (bHUD_exists) {
		Super::BuildHUDLines();
	}

	// 39(C) Add additional HUD lines for this model ---------------------
	line.message = FString::Printf(TEXT("dispZ = %f"), dispZ);
	line.colour = VARIBLE_COLOR;
	line.value = 0;
	HUDLines.Add(line);

	line.message = FString::Printf(TEXT("velyZ = %f"), velyZ);
	line.colour = VARIBLE_COLOR;
	line.value = 0;
	HUDLines.Add(line);
	// -------------------------------------------------------------------

}

void AMAS22_UTube::LogData() {
	int m;
	m = (int(logInterval / dT));
	if (itn % (m + 1) == 0) {		// m changed to m+1
		dataRecords.Add(FDataRecord_UTube{ time, dispZ, velyZ });
	}
}


// This is the main 'loop' function called on an Unreal Timer ------------
// Don't add any model-specific code in here
void AMAS22_UTube::ComputeDT() {
	if (bRunning) {
		LogData();					// 06-07-22 Moved to capture initial conds
		Computation();
		Visualization(dT);
		BuildHUDLines();
		itn++;
	}
}

void AMAS22_UTube::Computation() {

	switch (solver) {
	case SolverEnum::EULER:
		Euler(dT);
		break;
	case SolverEnum::RUNGE_KUTTA:
		RK(dT);
	case SolverEnum::RUNGE_KUTTA_FEHLBERG:
		RKF(dT);
		break;
	default:
		break;
	}

	// Addit2 (C) set model variables from u[] array ---------------------
	dispZ = u[0];
	velyZ = u[1];

	// -------------------------------------------------------------------
	// 30(C) Output computations for debugging ---------------------------
	//printf("time = %f, dispZ = %f, velyZ = %f\n", time, dispZ, velyZ);
	// -------------------------------------------------------------------

}

// Numerical solver code -------------------------------------------------

void AMAS22_UTube::Euler(double deltaT) {

}

void AMAS22_UTube::RK(double deltaT) {
	// Following for the Runge-Kutta experimental ------------------------
	double h = deltaT;
	flag = solverRK->step(u, dudt, time, h, yout, *rightHandSide);
	for (int i = 0; i < NREQUNS; i++) u[i] = yout[i];
	time += deltaT;
}

void AMAS22_UTube::RKF(double deltaT) {

	tout = time + deltaT;

	//printf(">>>> PRE  time = %f, tout = %f", time, tout);

	flag = solverRKF->r8_rkf45(NREQUNS, u, dudt, &time, tout, &relerr, abserr, flag, *rightHandSide);

	//printf(">>>> POST time = %f, tout = %f", time, tout);

	if (flag == 3) {
		printErr("Flag = 3 reset");
		flag = 2;
	}
	if (flag == 4) {
		//	printErr("Flag = 4 reset");
		flag = 2;
	}
	if (flag == 5) {
		printErr("Flag = 5 exiting");
		exit(0);
	}
	if (flag == 6) {
		printErr("Flag = 6 reset");
		flag = 2;
	}
	if (flag == 7) {
		//	printErr("Flag = 7 reset");
		flag = 2;
	}

}


void AMAS22_UTube::Visualization(float deltaTime) { // -------------------

	FVector scaleVec;
	scaleVec = LeftColumn->GetRelativeScale3D();
	scaleVec.Z = 1 + dispZ;
	LeftColumn->SetRelativeScale3D(scaleVec);

	scaleVec = RightColumn->GetRelativeScale3D();
	scaleVec.Z = 1 - dispZ;
	RightColumn->SetRelativeScale3D(scaleVec);

}

// Console commands ------------------------------------------------------

void AMAS22_UTube::SetConsoleAddit(FString name, FString strVal) {

}

void AMAS22_UTube::DumpConsoleAll() {
	Super::DumpConsoleAll();

}

// Create Octave file here -----------------------------------------------

void AMAS22_UTube::EndPlay(const EEndPlayReason::Type EndPlayReason) {
	Super::EndPlay(EndPlayReason);

	if (solver == SolverEnum::RUNGE_KUTTA) solverRK->cleanup();

	// Here comes the Octave code -------------
	
	FString logFile = FPaths::ProjectDir();
	FString fileName = "OctaveFiles/";
	fileName.Append(actorName);
	fileName.Append(".m");
	logFile.Append(fileName);

	
	int nrRecords = dataRecords.Num();

	printf("EndPlay() nrRecords = %d\n", nrRecords);

	FString stringToSave = "%col1 time, col2 dispZ, col3 velyZ\n\n";

	// Declare the Octave matrix "A" -------------------------------------
	stringToSave.Append("A=[\n");

	// Add the data from dataRecords[...] into array "A" -----------------

	for (int i = 0; i < nrRecords; i++) {
		stringToSave.Append(FString::Printf(TEXT("%e,  "), dataRecords[i].time));
		stringToSave.Append(FString::Printf(TEXT("%e,  "), dataRecords[i].dispZ));
		stringToSave.Append(FString::Printf(TEXT("%e,\n"), dataRecords[i].velyZ));
	}

	// Close the Octave matrix "A" ---------------------------------------
	stringToSave.Append("];\n\n");

	// Create some Octave variable names, from matrix "A" ----------------

	stringToSave.Append("time   = A(:,1);\n");
	stringToSave.Append("dispZ  = A(:,2);\n");
	stringToSave.Append("velyZ  = A(:,3);\n");

	// Get the parameters into Octave ------------------------------------

	stringToSave.Append(logf("gravity = %f, totalLength = %f, frictionCoeff = %f\n", gravityP, totalLengthP, frictionCoeffP));
	stringToSave.Append(logf("initDispZ = %f, initVelyZ = %f\n", initDispZ, initVelyZ));
	stringToSave.Append("if(initDispZ == 0.0) initDispZ = 1e-10; endif;\n");

	// calc the theoretical solution --------------------------------------

	// Here comes a cludge where we mis-appropriate stuff from the MK-AirTrack
	stringToSave.Append("m=1;\n");
	stringToSave.Append("b=frictionCoeff;\n");
	stringToSave.Append("k=2*gravity/totalLength;\n");
	
	stringToSave.Append("tau = 2*m/(b + 1e-7);\n");
	stringToSave.Append("omega = sqrt(k/m - (b/(2*m))^2);\n");
	stringToSave.Append("t=initDispZ^2 + (1/omega^2)*(initVelyZ + initDispZ/tau)^2;\n");
	stringToSave.Append(" A = sqrt(t);\n");
	stringToSave.Append("phi = -asin( (initVelyZ + initDispZ/tau)/(omega*A));;\n");
	stringToSave.Append("phi = -atan((1/tau + initVelyZ/initDispZ)/omega);\n");

	stringToSave.Append("xTh = A*exp(-time/tau).*cos(omega*time + phi);\n");
	stringToSave.Append("vTh = -(1/tau)*A*exp(-time/tau).*cos(omega*time+phi) - ...;\n");
	stringToSave.Append("A*exp(-time/tau)*omega.*sin(omega*time + phi);\n");

	// Plot experimental and theoretical solution ------------------------
	// First displacement ------------------------------------------------

	stringToSave.Append("figure(1)\n");
	stringToSave.Append("plot(time,dispZ,'o',time,xTh,'linewidth',3);\n");
	stringToSave.Append(logf("title('gravity = %f, totalLength = %f, frictionCoeff = %f');\n", gravityP, totalLengthP, frictionCoeffP));
	stringToSave.Append("xlabel('time')\n");
	stringToSave.Append("ylabel('dispZ')\n");
	stringToSave.Append(logf("grid;\n"));

	// Then velocity -----------------------------------------------------

	stringToSave.Append("figure(2)\n");
	stringToSave.Append("plot(time,velyZ,'o',time,vTh,'linewidth',3);\n");
	stringToSave.Append(logf("title('gravity = %f, totalLength = %f, frictionCoeff = %f');\n", gravityP, totalLengthP, frictionCoeffP));
	stringToSave.Append("xlabel('time')\n");
	stringToSave.Append("ylabel('velyZ')\n");
	stringToSave.Append(logf("grid;\n"));

	// All done. Write file to disk -------------------------------------

	FFileHelper::SaveStringToFile(stringToSave, *logFile);
	

}

void AMAS22_UTube::ApplyConfig() {
	// config file stuff --------------------------
	if (editorOverride == true) {
		FString configFName = FPaths::GeneratedConfigDir().Append(TEXT("MAS22_System.ini"));
		//printf("AMAS22_UTube::ActorInitialize() path to config file = %s", *configFName);
		GConfig->LoadFile(configFName);

		// System defaults -------------------------------------------------------------------------

		GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_UTube"), TEXT("scaling"), scaling, configFName);
		printf("AMAS22_UTube::ActorInitialize() >>>>>Config File read: scaling set to %f", scaling);

		GConfig->GetString(TEXT("/Script/MAS22_SciencePark.MAS22_UTube"), TEXT("defaultSolver"), defaultSolver, configFName);
		printf("AMAS22_UTube::ActorInitialize() >>>>>Config File read: solver set to %s", *defaultSolver);
		if (defaultSolver == "RUNGE_KUTTA_FEHLBERG") solver = SolverEnum::RUNGE_KUTTA_FEHLBERG;

		GConfig->GetString(TEXT("/Script/MAS22_SciencePark.MAS22_UTube"), TEXT("actorName"), actorName, configFName);
		printf("AMAS22_UTube::ActorInitialize() >>>>>Config File read: actor Name set to %s", *actorName);

		GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_UTube"), TEXT("dT"), dT, configFName);
		printf("AMAS22_UTube::ActorInitialize() >>>>>Config File read: dT set to %f", dT);

		// Default parameters ----------------------------------------------------------------------

		GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_UTube"), TEXT("gravity"), gravity, configFName);
		printf("AMAS22_UTube::ActorInitialize() >>>>>Config File read: gravity set to %f", gravity);

		GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_UTube"), TEXT("totalLength"), totalLength, configFName);
		printf("AMAS22_UTube::ActorInitialize() >>>>>Config File read: totalLength set to %f", totalLength);

		GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_UTube"), TEXT("frictionCoeff"), frictionCoeff, configFName);
		printf("AMAS22_UTube::ActorInitialize() >>>>>Config File read: frictionCoeff set to %f", frictionCoeff);
	}
}
