// =======================================================================
// CBP + FBO = MAS22 Simulation Framework
// MAS22_UTube.h
// Template for Matt Pethybridge
// 01-07-22
// 07-07-22
// =======================================================================

#pragma once

#include "CoreMinimal.h"
#include "MAS22_Actor.h"
#include "MAS22_Solver_RK.h"
#include "MAS22_Solver_RKF.h"
#include "MAS22_Structs.h"
#include "GameFramework/Actor.h"
#include "MAS22_UTube.generated.h"

// Right hand side Class capturing model dynamic system ------------------

class rhs_UTube {

public:

	double gravity, totalLength, frictionCoeff;

	rhs_UTube(double _gravity, double _totalLength, double _frictionCoeff) :
		gravity(_gravity), totalLength(_totalLength), frictionCoeff(_frictionCoeff)
	{}


	// System of ode's

	void operator() (double t, double u[], double dudt[]) {
		// 19(H) Associate state u[] with named variables----------------
		dispZ = u[0];
		velyZ = u[1];
		// --------------------------------------------------------------

		// 20(H) code the system of ODEs --------------------------------
		dudt[0] = velyZ;
		dudt[1] = -2.0 * (gravity / totalLength) * dispZ - (frictionCoeff) * velyZ;
		// -------------------------------------------------------------=
	}

private:
	double dispZ;
	double velyZ;

}; // end of rhs Class	--------------------------------------------------


// Structures specific to this model -------------------------------------
USTRUCT()
struct FDataRecord_UTube {
	GENERATED_BODY()
		double time, dispZ, velyZ;  // Specific to this model
};

UCLASS()
class MAS22_SCIENCEPARK_API AMAS22_UTube : public AMAS22_Actor
{
	GENERATED_BODY()

		UPROPERTY(VisibleDefaultsOnly)
		class UCapsuleComponent* CapsuleCollider;

	// 2(H) Add variable for the track --------------------------------------
	UPROPERTY(VisibleDefaultsOnly)
		class UStaticMeshComponent* UTube;

	// 4(H) Add variable for the glider -------------------------------------
	UPROPERTY(VisibleDefaultsOnly)
		class UStaticMeshComponent* ULiquid;

	// 6(H) Add variable for the left spring --------------------------------
	UPROPERTY(VisibleDefaultsOnly)
		class UStaticMeshComponent* LeftColumn;

	// 9(H) Add variable for the right spring --------------------------------
	UPROPERTY(VisibleDefaultsOnly)
		class UStaticMeshComponent* RightColumn;

	// 11(H) Add variable for the left stop--------------------------------
	UPROPERTY(VisibleDefaultsOnly)
		class UStaticMeshComponent* Stand;

	// 14(H) Add variable for the right stop--------------------------------
	UPROPERTY(VisibleDefaultsOnly)
		class UStaticMeshComponent* Clamp;

public:

	AMAS22_UTube();

private:

	// Parameters -----------------------------------------------------------
	// variables used within the .cpp file. Append P for parameter so that
	// actual names can appear in Editable variables.

	double gravityP, totalLengthP, frictionCoeffP;

public:

	// Editable variables ---------------------------------------------------

	//UPROPERTY(Config, EditAnywhere)
	//	double scaling = 100.0; // 100 Unreal Units per metre

	// 21(H) Declare parameters and initial conditions so they appear in editor
	// These are 'deafault' variables which appear in config
	UPROPERTY(EditAnywhere, Config)
		double gravity = 9.8;

	UPROPERTY(EditAnywhere, Config)
		double totalLength = 1.0;

	UPROPERTY(EditAnywhere, Config)
		double frictionCoeff = 0.0;

	// Initial conditions ---------------------------------------------------

	UPROPERTY(EditAnywhere)
		double initDispZ = 0.5;

	UPROPERTY(EditAnywhere)
		double initVelyZ = 0;
	// -------------------------------------------------------------------


protected:

	// Overriden functions from MAS22_Actor ------------------------------

	virtual void Computation()  override;
	virtual void Visualization(float deltaTime)  override;
	virtual void BuildHUDLines()  override;
	virtual void LogData()  override;
	virtual void EndPlay(const EEndPlayReason::Type EndPlayReason) override;
	virtual void ApplyConfig() override;

	void Euler(double dT)  override;
	void RKF(double dT)  override;
	void RK(double dT)  override;

	virtual void ActorInitialize() override;
	virtual void InitializeSolver() override;
	virtual void ComputeDT() override;

	virtual void SetModelParameters() override;
	virtual void UpdateModelParameters() override;
	virtual void GetDefaultParameters() override;

	virtual void SetModelICs() override;
	virtual void UpdateModelICs() override;

	virtual void DumpConsoleAll() override;

	virtual void SetConsoleAddit(FString name, FString strVal) override;

	// ----------------------------------------------------------------------

	// Additional variables required by the model, but not related to the
	// model dynamics.
	// 
	// 31(H) declare the glider start location ---------------------------
	FVector gliderStartLocation;
	// ------------------------------------------------------------------
	// 34(H) declare both springs start location -------------------------
	FVector leftSpringStartLocation;
	FVector rightSpringStartLocation;
	// -------------------------------------------------------------------
	// 37(H) Declare the spring length -----------------------------------
	double springLength = 220;
	// -------------------------------------------------------------------

public:
	rhs_UTube* rightHandSide;
	MAS22_Solver_RK* solverRK;
	MAS22_Solver_RKF* solverRKF;

private:
	// Addit1(H) Model variables to be used in cpp --------------------------
	double dispZ;
	double velyZ;
	// ----------------------------------------------------------------------

	// TArray for data logging ----------------------------------------------
	TArray<FDataRecord_UTube> dataRecords;

};

